import { createEslintRule } from '../utils/create-eslint-rule.mjs'
import { getGroupNumber } from '../utils/get-group-number.mjs'
import { getSourceCode } from '../utils/get-source-code.mjs'
import { toSingleLine } from '../utils/to-single-line.mjs'
import { rangeToDiff } from '../utils/range-to-diff.mjs'
import { getSettings } from '../utils/get-settings.mjs'
import { isPositive } from '../utils/is-positive.mjs'
import { sortNodes } from '../utils/sort-nodes.mjs'
import { useGroups } from '../utils/use-groups.mjs'
import { makeFixes } from '../utils/make-fixes.mjs'
import { complete } from '../utils/complete.mjs'
import { pairwise } from '../utils/pairwise.mjs'
import { compare } from '../utils/compare.mjs'
const sortIntersectionTypes = createEslintRule({
  name: 'sort-intersection-types',
  meta: {
    type: 'suggestion',
    docs: {
      description: 'Enforce sorted intersection types.',
    },
    fixable: 'code',
    schema: [
      {
        type: 'object',
        properties: {
          type: {
            description: 'Specifies the sorting method.',
            type: 'string',
            enum: ['alphabetical', 'natural', 'line-length'],
          },
          order: {
            description:
              'Determines whether the sorted items should be in ascending or descending order.',
            type: 'string',
            enum: ['asc', 'desc'],
          },
          ignoreCase: {
            description:
              'Controls whether sorting should be case-sensitive or not.',
            type: 'boolean',
          },
          groups: {
            description: 'Specifies the order of the groups.',
            type: 'array',
            items: {
              oneOf: [
                {
                  type: 'string',
                },
                {
                  type: 'array',
                  items: {
                    type: 'string',
                  },
                },
              ],
            },
          },
        },
        additionalProperties: false,
      },
    ],
    messages: {
      unexpectedIntersectionTypesOrder:
        'Expected "{{right}}" to come before "{{left}}".',
    },
  },
  defaultOptions: [
    {
      type: 'alphabetical',
      order: 'asc',
      ignoreCase: true,
      groups: [],
    },
  ],
  create: context => ({
    TSIntersectionType: node => {
      let settings = getSettings(context.settings)
      let options = complete(context.options.at(0), settings, {
        type: 'alphabetical',
        ignoreCase: true,
        order: 'asc',
        groups: [],
      })
      let sourceCode = getSourceCode(context)
      let nodes = node.types.map(type => {
        let { getGroup, defineGroup } = useGroups(options.groups)
        switch (type.type) {
          case 'TSConditionalType':
            defineGroup('conditional')
            break
          case 'TSConstructorType':
          case 'TSFunctionType':
            defineGroup('function')
            break
          case 'TSImportType':
            defineGroup('import')
            break
          case 'TSIntersectionType':
            defineGroup('intersection')
            break
          case 'TSAnyKeyword':
          case 'TSBigIntKeyword':
          case 'TSBooleanKeyword':
          case 'TSNeverKeyword':
          case 'TSNumberKeyword':
          case 'TSObjectKeyword':
          case 'TSStringKeyword':
          case 'TSSymbolKeyword':
          case 'TSThisType':
          case 'TSUnknownKeyword':
          case 'TSIntrinsicKeyword':
            defineGroup('keyword')
            break
          case 'TSLiteralType':
          case 'TSTemplateLiteralType':
            defineGroup('literal')
            break
          case 'TSArrayType':
          case 'TSIndexedAccessType':
          case 'TSInferType':
          case 'TSTypeReference':
          case 'TSQualifiedName':
            defineGroup('named')
            break
          case 'TSMappedType':
          case 'TSTypeLiteral':
            defineGroup('object')
            break
          case 'TSTypeQuery':
          case 'TSTypeOperator':
            defineGroup('operator')
            break
          case 'TSTupleType':
            defineGroup('tuple')
            break
          case 'TSUnionType':
            defineGroup('union')
            break
          case 'TSNullKeyword':
          case 'TSUndefinedKeyword':
          case 'TSVoidKeyword':
            defineGroup('nullish')
            break
        }
        return {
          name: sourceCode.text.slice(...type.range),
          size: rangeToDiff(type.range),
          group: getGroup(),
          node: type,
        }
      })
      pairwise(nodes, (left, right) => {
        let leftNum = getGroupNumber(options.groups, left)
        let rightNum = getGroupNumber(options.groups, right)
        if (
          leftNum > rightNum ||
          (leftNum === rightNum && isPositive(compare(left, right, options)))
        ) {
          context.report({
            messageId: 'unexpectedIntersectionTypesOrder',
            data: {
              left: toSingleLine(left.name),
              right: toSingleLine(right.name),
            },
            node: right.node,
            fix: fixer => {
              let grouped = {}
              for (let currentNode of nodes) {
                let groupNum = getGroupNumber(options.groups, currentNode)
                if (!(groupNum in grouped)) {
                  grouped[groupNum] = [currentNode]
                } else {
                  grouped[groupNum] = sortNodes(
                    [...grouped[groupNum], currentNode],
                    options,
                  )
                }
              }
              let sortedNodes = []
              for (let group of Object.keys(grouped).sort(
                (a, b) => Number(a) - Number(b),
              )) {
                sortedNodes.push(...sortNodes(grouped[group], options))
              }
              return makeFixes(fixer, nodes, sortedNodes, sourceCode)
            },
          })
        }
      })
    },
  }),
})
export { sortIntersectionTypes as default }
